# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import sys
from . import main

def entry_point():
   return main.main(sys.argv)
